@@ -21,6 +21,7 @@ from typing import Optional, Union, Sequence
 import jax
 import jax.numpy as jnp
 from ott.geometry import epsilon_scheduler
+from ott.geometry import ops
 
 
 @jax.tree_util.register_pytree_node_class
@@ -156,9 +157,8 @@ class Geometry:
     """
     h = (f if axis == 1 else g)
     z = self.apply_lse_kernel(f, g, self.epsilon, axis=axis)[0]
-    return jnp.where(jnp.isfinite(h),
-                     jnp.exp((z + h)/self.epsilon),
-                     0)
+    return jnp.exp((z + h)/self.epsilon)
+
 
   def marginal_from_scalings(self, u: jnp.ndarray, v: jnp.ndarray, axis=0):
     """Outputs marginal of transportation matrix from scalings."""
@@ -200,7 +200,7 @@ class Geometry:
       marginal = self.marginal_from_scalings(f_u, g_v, axis=axis)
     norm_error = jnp.array(norm_error)
     error = jnp.sum(
-        jnp.abs(marginal - target)**norm_error[:, jnp.newaxis],
+        jnp.abs(marginal - target) ** norm_error[:, jnp.newaxis],
         axis=1) ** (1.0 / norm_error)
     return error
 
@@ -247,11 +247,14 @@ class Geometry:
     if vec is not None:
       if axis == 0:
         vec = vec.reshape((vec.size, 1))
-      lse_output = jax.scipy.special.logsumexp(
-          self._center(f, g) / eps, b=vec, axis=axis, return_sign=True)
+      lse_output = ops.logsumexp(
+          self._center(f, g) / eps,
+          b=vec,
+          axis=axis,
+          return_sign=True)
       return eps * lse_output[0], lse_output[1]
     else:
-      lse_output = jax.scipy.special.logsumexp(
+      lse_output = ops.logsumexp(
           self._center(f, g) / eps, axis=axis, return_sign=False)
       return eps * lse_output, jnp.array([1.0])
 
@@ -326,30 +329,51 @@ class Geometry:
     return self.epsilon * jnp.log(scaling)
 
   def scaling_from_potential(self, potential: jnp.ndarray) -> jnp.ndarray:
-    return jnp.exp(potential / self.epsilon)
+    finite = jnp.isfinite(potential)
+    return jnp.where(finite,
+                     jnp.exp(jnp.where(finite, potential / self.epsilon, 0.0)),
+                     0.0)
 
-  def apply_cost(self, arr: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
+  def apply_cost(self, arr: jnp.ndarray, axis: int = 0, fn=None) -> jnp.ndarray:
     """Applies cost matrix to array (vector or matrix).
 
+    This function applies the ground geometry's cost matrix, to perform either
+    output = K arr (if axis=1)
+    output = K' arr (if axis=0)
+    where K is [num_a, num_b]
+
     Args:
-      arr: jnp.ndarray [num_b,...] or [num_a,...], depending on axis.
-      axis: axis.
+      arr: jnp.ndarray [num_a or num_b, p], vector that will be multiplied
+        by the cost matrix.
+      axis: standard cost matrix if axis=1, transport if 0
+      fn: function to apply to cost matrix element-wise before the dot product
+
     Returns:
-      A jnp.ndarray corresponding to cost x matrix
+      A jnp.ndarray, [num_b, p] if axis=0 or [num_a, p] if axis=1
     """
-    return jax.vmap(lambda x: self._apply_cost_to_vec(x, axis)
-                   )(jnp.atleast_2d(arr))
+    if arr.ndim == 1:
+      return jax.vmap(lambda x: self._apply_cost_to_vec(x, axis, fn), 1, 1,
+                     )(arr.reshape(-1, 1))
+    return jax.vmap(lambda x: self._apply_cost_to_vec(x, axis, fn), 1, 1,
+                   )(arr)
 
-  def _apply_cost_to_vec(self, vec: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
-    """Applies [num_a, num_b] cost matrix to vector.
+  def _apply_cost_to_vec(self,
+                         vec: jnp.ndarray,
+                         axis: int = 0,
+                         fn=None) -> jnp.ndarray:
+    """Applies [num_a, num_b] fn(cost) (or transpose) to vector.
 
     Args:
-      vec: jnp.ndarray [num_b,] ([num_a,] if axis=1) vector
-      axis: axis.
+      vec: jnp.ndarray [num_a,] ([num_b,] if axis=1) vector
+      axis: axis on which the reduction is done.
+      fn: function optionally applied to cost matrix element-wise, before the
+        doc product
     Returns:
       A jnp.ndarray corresponding to cost x vector
     """
-    return jnp.dot(self.cost_matrix if axis == 0 else self.cost_matrix.T, vec)
+    matrix = self.cost_matrix.T if axis == 0 else self.cost_matrix
+    matrix = fn(matrix) if fn is not None else matrix
+    return jnp.dot(matrix, vec)
 
   @classmethod
   def prepare_divergences(cls, *args, static_b: bool = False, **kwargs):
@@ -365,7 +389,7 @@ class Geometry:
     )
 
   def tree_flatten(self):
-    return (self.cost_matrix, self.kernel_matrix, self._epsilon), None
+    return (self._cost_matrix, self._kernel_matrix, self._epsilon), None
 
   @classmethod
   def tree_unflatten(cls, aux_data, children):
