@@ -24,6 +24,7 @@ import numpy as np
 from ott.geometry import costs
 from ott.geometry import epsilon_scheduler
 from ott.geometry import geometry
+from ott.geometry import ops
 
 
 @jax.tree_util.register_pytree_node_class
@@ -79,7 +80,8 @@ class Grid(geometry.Geometry):
       **kwargs: passed to parent class.
     """
     super().__init__(epsilon=epsilon, **kwargs)
-    if grid_size is not None and x is not None and num_a is not None and grid_dimension is not None:
+    if (grid_size is not None and x is not None and num_a is not None and
+        grid_dimension is not None):
       self.grid_size = grid_size
       self.x = x
       self.num_a = num_a
@@ -185,15 +187,18 @@ class Grid(geometry.Geometry):
 
     if vec is not None:
       vec = jnp.transpose(vec, indices)
-      softmax_res, softmax_sgn = jax.scipy.special.logsumexp(
+      softmax_res, softmax_sgn = ops.logsumexp(
           centered_cost, b=vec, axis=1, return_sign=True)
       return eps * jnp.transpose(softmax_res, indices), jnp.transpose(
           softmax_sgn, indices)
     else:
-      softmax_res = jax.scipy.special.logsumexp(centered_cost, axis=1)
+      softmax_res = ops.logsumexp(centered_cost, axis=1)
       return eps * jnp.transpose(softmax_res, indices), None
 
-  def _apply_cost_to_vec(self, vec: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
+  def _apply_cost_to_vec(self,
+                         vec: jnp.ndarray,
+                         axis: int = 0,
+                         fn=None) -> jnp.ndarray:
     r"""Applies grid's cost matrix (without instantiating it) to a vector.
 
     The `apply_cost` operation on grids rests on the following identity.
@@ -213,7 +218,9 @@ class Grid(geometry.Geometry):
 
     Args:
       vec: jnp.ndarray, flat vector of total size prod(grid_size).
-      axis: axis 0 if applying original costs, 1 if using their transpose.
+      axis: axis 0 if applying transpose costs, 1 if using the original cost.
+      fn: function optionally applied to cost matrix element-wise, before the
+        dot product.
 
     Returns:
       A jnp.ndarray corresponding to cost x matrix
@@ -224,7 +231,7 @@ class Grid(geometry.Geometry):
     for dimension, cost in enumerate(self.cost_matrices):
       ind = indices.copy()
       ind.insert(dimension, 0)
-      if axis == 1:
+      if axis == 0:
         cost = cost.T
       accum_vec += np.sum(
           jnp.tensordot(cost, vec, axes=([0], [dimension])),
